@@ -22,30 +22,22 @@
 package org.kapott.hbci.passport;
 
 import java.io.File;
-import java.io.FileInputStream;
-import java.io.FileOutputStream;
-import java.io.ObjectInputStream;
-import java.io.ObjectOutputStream;
-import java.io.StreamCorruptedException;
 import java.util.List;
 import java.util.Properties;
 import java.util.StringTokenizer;
 
-import javax.crypto.Cipher;
-import javax.crypto.CipherInputStream;
-import javax.crypto.CipherOutputStream;
-import javax.crypto.SecretKey;
-import javax.crypto.spec.PBEParameterSpec;
-
 import org.kapott.hbci.callback.HBCICallback;
 import org.kapott.hbci.exceptions.HBCI_Exception;
-import org.kapott.hbci.exceptions.InvalidPassphraseException;
 import org.kapott.hbci.manager.FlickerCode;
 import org.kapott.hbci.manager.HBCIUtils;
 import org.kapott.hbci.manager.HBCIUtilsInternal;
 import org.kapott.hbci.manager.HHDVersion;
 import org.kapott.hbci.manager.HHDVersion.Type;
 import org.kapott.hbci.manager.LogFilter;
+import org.kapott.hbci.manager.MatrixCode;
+import org.kapott.hbci.manager.QRCode;
+import org.kapott.hbci.passport.storage.PassportData;
+import org.kapott.hbci.passport.storage.PassportStorage;
 import org.kapott.hbci.security.Sig;
 
 /** <p>Passport-Klasse für HBCI mit PIN/TAN. Dieses Sicherheitsverfahren wird erst
@@ -62,28 +54,31 @@ import org.kapott.hbci.security.Sig;
     Transportschicht. Die Nachrichten werden nämlich nicht direkt via TCP/IP übertragen,
     sondern in das HTTP-Protokoll eingebettet. Die Verschlüsselung der übertragenen Daten
     erfolgt dabei auf HTTP-Ebene (via SSL = HTTPS).</p><p>
-    Wie auch bei {@link org.kapott.hbci.passport.HBCIPassportRDH} wird eine "Schlüsseldatei"
+    Wie auch bei {@link org.kapott.hbci.passport.HBCIPassportRDHNew} wird eine "Schlüsseldatei"
     verwendet. In dieser werden allerdings keine kryptografischen Schlüssel abgelegt, sondern
     lediglich die Zugangsdaten für den HBCI-Server (Hostadresse, Nutzerkennung, usw.) sowie
     einige zusätzliche Daten (BPD, UPD, zuletzt benutzte HBCI-Version). Diese Datei wird
     vor dem Abspeichern verschlüsselt. Vor dem Erzeugen bzw. erstmaligen Einlesen wird via
     Callback nach einem Passwort gefragt, aus welchem der Schlüssel für die Verschlüsselung
     der Datei berechnet wird</p>*/
-public class HBCIPassportPinTan
-    extends AbstractPinTanPassport
+public class HBCIPassportPinTan extends AbstractPinTanPassport
 {
-    private String    filename;
-    private SecretKey passportKey;
-
-    private final static byte[] CIPHER_SALT={(byte)0x26,(byte)0x19,(byte)0x38,(byte)0xa7,
-                                             (byte)0x99,(byte)0xbc,(byte)0xf1,(byte)0x55};
-    private final static int CIPHER_ITERATIONS=987;
+    private String filename;
 
+    /**
+     * ct.
+     * @param init
+     * @param dummy
+     */
     public HBCIPassportPinTan(Object init,int dummy)
     {
         super(init);
     }
 
+    /**
+     * ct.
+     * @param initObject
+     */
     public HBCIPassportPinTan(Object initObject)
     {
         this(initObject,0);
@@ -100,7 +95,8 @@ public class HBCIPassportPinTan
         setProxyUser(HBCIUtils.getParam(header+"proxyuser",""));
         setProxyPass(HBCIUtils.getParam(header+"proxypass",""));
 
-        if (init) {
+        if (init)
+        {
             this.read();
 
             
@@ -132,7 +128,6 @@ public class HBCIPassportPinTan
      */
     public void resetPassphrase()
     {
-        passportKey=null;
     }
     
     /**
@@ -165,86 +160,23 @@ public class HBCIPassportPinTan
         create();
         
         String fname = this.getFileName();
-        if (fname==null) {
+        if (fname == null)
             throw new NullPointerException("client.passport.PinTan.filename must not be null");
-        }
 
-        HBCIUtils.log("loading data from file " + fname,HBCIUtils.LOG_DEBUG);
-
-        ObjectInputStream o = null;
-        try
-        {
-            int retries = Integer.parseInt(HBCIUtils.getParam("client.retries.passphrase","3"));
-            
-            while (true) {
-                if (passportKey == null)
-                    passportKey = calculatePassportKey(FOR_LOAD);
-
-                PBEParameterSpec paramspec=new PBEParameterSpec(CIPHER_SALT,CIPHER_ITERATIONS);
-                String provider = HBCIUtils.getParam("kernel.security.provider");
-                Cipher cipher = provider == null ? Cipher.getInstance("PBEWithMD5AndDES") : Cipher.getInstance("PBEWithMD5AndDES", provider);
-                cipher.init(Cipher.DECRYPT_MODE,passportKey,paramspec);
-                
-                o = null;
-                try
-                {
-                    o=new ObjectInputStream(new CipherInputStream(new FileInputStream(fname),cipher));
-                }
-                catch (StreamCorruptedException e)
-                {
-                    passportKey=null;
-                    
-                    retries--;
-                    if (retries<=0)
-                        throw new InvalidPassphraseException();
-                }
-                
-                if (o!=null)
-                    break;
-            }
-
-            setCountry((String)(o.readObject()));
-            setBLZ((String)(o.readObject()));
-            setHost((String)(o.readObject()));
-            setPort((Integer)(o.readObject()));
-            setUserId((String)(o.readObject()));
-            setSysId((String)(o.readObject()));
-            setBPD((Properties)(o.readObject()));
-            setUPD((Properties)(o.readObject()));
-
-            setHBCIVersion((String)o.readObject());
-            setCustomerId((String)o.readObject());
-            setFilterType((String)o.readObject());
-            
-            try {
-                setAllowedTwostepMechanisms((List<String>)o.readObject());
-                try
-                {
-                    setCurrentTANMethod((String)o.readObject());
-                }
-                catch (Exception e)
-                {
-                    HBCIUtils.log("no current secmech found in passport file - automatically upgrading to new file format", HBCIUtils.LOG_WARN);
-                }
-            }
-            catch (Exception e)
-            {
-                HBCIUtils.log("no list of allowed secmechs found in passport file - automatically upgrading to new file format", HBCIUtils.LOG_WARN);
-            }
-        } 
-        catch (Exception e)
-        {
-            throw new HBCI_Exception("*** loading of passport file failed",e);
-        }
-
-        try
-        {
-            o.close();
-        }
-        catch (Exception e)
-        {
-            HBCIUtils.log(e);
-        }
+        PassportData data = PassportStorage.load(this,new File(fname));
+        this.setCountry(data.country);
+        this.setBLZ(data.blz);
+        this.setHost(data.host);
+        this.setPort(data.port);
+        this.setUserId(data.userId);
+        this.setSysId(data.sysId);
+        this.setBPD(data.bpd);
+        this.setUPD(data.upd);
+        this.setHBCIVersion(data.hbciVersion);
+        this.setCustomerId(data.customerId);
+        this.setFilterType(data.filter);
+        this.setAllowedTwostepMechanisms(data.twostepMechs);
+        this.setCurrentTANMethod(data.tanMethod);
     }
 
     /**
@@ -253,52 +185,32 @@ public class HBCIPassportPinTan
     @Override
     public void saveChanges()
     {
-        File passportfile = new File(getFileName());
-        File tempfile     = null;
-        
-        try
-        {
-            if (passportKey==null) 
-                passportKey=calculatePassportKey(FOR_SAVE);
+        try {
+
+            final PassportData data = new PassportData();
             
-            PBEParameterSpec paramspec=new PBEParameterSpec(CIPHER_SALT,CIPHER_ITERATIONS);
-            String provider = HBCIUtils.getParam("kernel.security.provider");
-            Cipher cipher = provider == null ? Cipher.getInstance("PBEWithMD5AndDES") : Cipher.getInstance("PBEWithMD5AndDES", provider);
-            cipher.init(Cipher.ENCRYPT_MODE,passportKey,paramspec);
+            data.country     = this.getCountry();
+            data.blz         = this.getBLZ();
+            data.host        = this.getHost();
+            data.port        = this.getPort();
+            data.userId      = this.getUserId();
+            data.sysId       = this.getSysId();
+            data.bpd         = this.getBPD();
+            data.upd         = this.getUPD();
 
-            File directory = passportfile.getAbsoluteFile().getParentFile();
-            String prefix  = passportfile.getName()+"_";
-            tempfile       = File.createTempFile(prefix,"",directory);
-
-            HBCIUtils.log("writing to passport file " + tempfile, HBCIUtils.LOG_DEBUG);
-            ObjectOutputStream o = new ObjectOutputStream(new CipherOutputStream(new FileOutputStream(tempfile),cipher));
-
-            o.writeObject(getCountry());
-            o.writeObject(getBLZ());
-            o.writeObject(getHost());
-            o.writeObject(getPort());
-            o.writeObject(getUserId());
-            o.writeObject(getSysId());
-            o.writeObject(getBPD());
-            o.writeObject(getUPD());
-
-            o.writeObject(getHBCIVersion());
-            o.writeObject(getCustomerId());
-            o.writeObject(getFilterType());
+            data.hbciVersion = this.getHBCIVersion();
+            data.customerId  = this.getCustomerId();
+            data.filter      = this.getFilterType();
             
-            // hier auch gewähltes zweischritt-verfahren abspeichern
-            List<String> l = getAllowedTwostepMechanisms();
+            final List<String> l = getAllowedTwostepMechanisms();
             HBCIUtils.log("saving two step mechs: " + l, HBCIUtils.LOG_DEBUG);
-            o.writeObject(l);
+            data.twostepMechs = l;
             
-            String s=getCurrentTANMethod(false);
+            final String s = this.getCurrentTANMethod(false);
             HBCIUtils.log("saving current tan method: "+s, HBCIUtils.LOG_DEBUG);
-            o.writeObject(s);
+            data.tanMethod = s;
 
-            HBCIUtils.log("closing output stream", HBCIUtils.LOG_DEBUG);
-            o.close();
-            
-            this.safeReplace(passportfile,tempfile);
+            PassportStorage.save(this,data,new File(this.getFileName()));
         }
         catch (HBCI_Exception he)
         {
@@ -306,7 +218,7 @@ public class HBCIPassportPinTan
         }
         catch (Exception e)
         {
-            throw new HBCI_Exception("*** saving of passport file failed",e);
+            throw new HBCI_Exception(HBCIUtilsInternal.getLocMsg("EXCMSG_PASSPORT_WRITEERR"),e);
         }
     }
     
@@ -431,34 +343,42 @@ public class HBCIPassportPinTan
 
                     HHDVersion hhd = HHDVersion.find(secmechInfo);
                     HBCIUtils.log("detected HHD version: " + hhd,HBCIUtils.LOG_DEBUG);
-
+                    
                     final StringBuffer payload = new StringBuffer();
                     final String msg = secmechInfo.getProperty("name")+"\n"+secmechInfo.getProperty("inputinfo")+"\n\n"+challenge;
-                    
-                    if (hhd.getType() == Type.PHOTOTAN)
+
+                    int callback = HBCICallback.NEED_PT_TAN;
+
+                    // Um sicherzustellen, dass wird keinen falschen Callback ausloesen, weil wir die HHD-Version
+                    // eventuell falsch erkannt haben, versuchen wir bei PhotoTAN und QR-Code zusaetzlich, die Daten
+                    // zu parsen. Nur wenn sie korrekt geparst werden koennen, verwenden wir auch den spezifischen Callback
+                    if (hhd.getType() == Type.PHOTOTAN && (MatrixCode.tryParse(hhduc) != null))
                     {
                         // Bei PhotoTAN haengen wir ungeparst das HHDuc an. Das kann dann auf
                         // Anwendungsseite per MatrixCode geparst werden
                         payload.append(hhduc);
-                        HBCIUtilsInternal.getCallback().callback(this,HBCICallback.NEED_PT_PHOTOTAN,msg,HBCICallback.TYPE_TEXT,payload);
+                        callback = HBCICallback.NEED_PT_PHOTOTAN;
                     }
-                    else if (hhd.getType() == Type.QRCODE)
+                    else if (hhd.getType() == Type.QRCODE && (QRCode.tryParse(hhduc,msg) != null))
                     {
                         // Bei QR-Code haengen wir ungeparst das HHDuc an. Das kann dann auf
                         // Anwendungsseite per QRCode geparst werden
                         payload.append(hhduc);
-                        HBCIUtilsInternal.getCallback().callback(this,HBCICallback.NEED_PT_QRTAN,msg,HBCICallback.TYPE_TEXT,payload);
+                        callback = HBCICallback.NEED_PT_QRTAN;
                     }
                     else
                     {
-                        // willuhn 2011-05-27: Flicker-Code anhaengen, falls vorhanden
-                        String flicker = parseFlickercode(challenge,hhduc);
+                        FlickerCode flicker = FlickerCode.tryParse(challenge,hhduc);
                         if (flicker != null)
-                            payload.append(flicker);
-                        
-                        HBCIUtilsInternal.getCallback().callback(this,HBCICallback.NEED_PT_TAN,msg,HBCICallback.TYPE_TEXT,payload);
+                        {
+                            // Bei chipTAN liefern wir den bereits geparsten und gerenderten Flickercode
+                            payload.append(flicker.render());
+                        }
                     }
 
+                    // Callback durchfuehren
+                    HBCIUtilsInternal.getCallback().callback(this,callback,msg,HBCICallback.TYPE_TEXT,payload);
+
                     setPersistentData("externalid",null); // External-ID aus Passport entfernen
                     if (payload == null || payload.length()==0) {
                         throw new HBCI_Exception(HBCIUtilsInternal.getLocMsg("EXCMSG_TANZERO"));
@@ -477,63 +397,17 @@ public class HBCIPassportPinTan
     }
     
     /**
-     * Versucht, aus Challenge und Challenge HHDuc den Flicker-Code zu extrahieren
-     * und ihn in einen flickerfaehigen Code umzuwandeln.
-     * Nur wenn tatsaechlich ein gueltiger Code enthalten ist, der als
-     * HHDuc-Code geparst und in einen Flicker-Code umgewandelt werden konnte,
-     * liefert die Funktion den Code. Sonst immer NULL.
-     * @param challenge der Challenge-Text. Das DE "Challenge HHDuc" gibt es
-     * erst seit HITAN4. Einige Banken haben aber schon vorher optisches chipTAN
-     * gemacht. Die haben das HHDuc dann direkt im Freitext des Challenge
-     * mitgeschickt (mit String-Tokens zum Extrahieren markiert). Die werden vom
-     * FlickerCode-Parser auch unterstuetzt.
-     * @param hhduc das echte Challenge HHDuc.
-     * @return der geparste und in Flicker-Format konvertierte Code oder NULL.
+     * @see org.kapott.hbci.passport.HBCIPassportInternal#verify(byte[], byte[])
      */
-    private String parseFlickercode(String challenge, String hhduc)
-    {
-      // 1. Prioritaet hat hhduc. Gibts aber erst seit HITAN4
-      if (hhduc != null && hhduc.trim().length() > 0)
-      {
-        try
-        {
-          FlickerCode code = new FlickerCode(hhduc);
-          return code.render();
-        }
-        catch (Exception e)
-        {
-          HBCIUtils.log("unable to parse Challenge HHDuc " + hhduc + ":" + HBCIUtils.exception2String(e),HBCIUtils.LOG_DEBUG);
-        }
-      }
-      
-      // 2. Checken, ob im Freitext-Challenge was parse-faehiges steht.
-      // Kann seit HITAN1 auftreten
-      if (challenge != null && challenge.trim().length() > 0)
-      {
-        try
-        {
-          FlickerCode code = new FlickerCode(challenge);
-          return code.render();
-        }
-        catch (Exception e)
-        {
-          // Das darf durchaus vorkommen, weil das Challenge auch bei manuellem
-          // chipTAN- und smsTAN Verfahren verwendet wird, wo gar kein Flicker-Code enthalten ist.
-          // Wir loggen es aber trotzdem - fuer den Fall, dass tatsaechlich ein Flicker-Code
-          // enthalten ist. Sonst koennen wir das nicht debuggen.
-          HBCIUtils.log("challenge contains no HHDuc (no problem in most cases):" + HBCIUtils.exception2String(e),HBCIUtils.LOG_DEBUG2);
-        }
-      }
-      // Ne, definitiv kein Flicker-Code.
-      return null;
-    }
-
     public boolean verify(byte[] data,byte[] sig)
     {
         // TODO: fuer bankensignaturen fuer HITAN muss dass hier geändert werden
         return true;
     }
 
+    /**
+     * @see org.kapott.hbci.passport.HBCIPassportInternal#encrypt(byte[])
+     */
     public byte[][] encrypt(byte[] plainMsg)
     {
         try {
@@ -545,6 +419,9 @@ public class HBCIPassportPinTan
         }
     }
 
+    /**
+     * @see org.kapott.hbci.passport.HBCIPassportInternal#decrypt(byte[], byte[])
+     */
     public byte[] decrypt(byte[] cryptedKey,byte[] cryptedMsg)
     {
         try {
@@ -554,10 +431,12 @@ public class HBCIPassportPinTan
         }
     }
     
+    /**
+     * @see org.kapott.hbci.passport.AbstractHBCIPassport#close()
+     */
     public void close()
     {
         super.close();
-        passportKey=null;
     }
     
 }
